import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import sun.tools.tree.ThisExpression;

public class PictViewer extends Frame implements ActionListener, MouseListener,
		MouseMotionListener, MouseWheelListener, KeyListener, WindowListener {
	
	private static final long serialVersionUID = 42L;
	int lastX=0, lastY=0;
	int pressedX, pressedY;
	int offsetX=0, offsetY=0;
	int imageAngle=0;
	boolean isPenDown=false;
	boolean isPaperweightOn=false;
	boolean wasPaperweightOnWhenPressed=false;
	boolean isDoodleDemo=true;
	boolean isMapDemo, isPictureDemo;
	MenuBar mb;
	Menu mDemo, mColor, mPicture;
	Color currentColor = Color.black; //color of the pen
	final int SIZEX=480, SIZEY=640; //initial size of the screen
//	final int SIZEX=480, SIZEY=480; //initial size of the screen
//	int  pictureSizeX=2850, pictureSizeY=1542; //size of the picture
//	int  pictureSizeX=1280, pictureSizeY=1280; //size of the picture
	int  pictureSizeX, pictureSizeY; //size of the picture
	double zoomRatio = 1.0; //zooming factor, 1=real size
	Image bufferImage = null, pictureImage =null;
	Graphics bufferGraphics = null;
	
	public PictViewer () {
//	make manu bar
		mb = new MenuBar();
		mb.add(mDemo = new Menu("Demo"));
		mDemo.add("MapDemo");
		mDemo.add("PictureDemo");
		mDemo.addSeparator();
		mDemo.add("Doodle");
		mDemo.add("ZoomRotate");
		mDemo.addActionListener(this);
		mb.add(mColor = new Menu("Color"));
		mColor.add("Black");
		mColor.add("Red");
		mColor.add("Green");
		mColor.add("Blue");
		mColor.add("Yellow");
		mColor.add("White");
		mColor.addActionListener(this);
		mb.add(mPicture = new Menu("Picture"));
		mPicture.add("Rotate CW");
		mPicture.add("Rotate CCW");
		mPicture.addSeparator();
		mPicture.add("Zoom x0.1");
		mPicture.add("Zoom x0.5");
		mPicture.add("Zoom x1");
		mPicture.add("Zoom x2");
		mPicture.add("Zoom x4");
		mPicture.addActionListener(this);
		setMenuBar(mb);
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
		addKeyListener(this);
		addWindowListener(this);
		setSize(SIZEX,SIZEY);
//		setUndecorated(true); //remove menu bar
//		setResizable(false);
		setVisible(true);
		
//	prepare picture
		this.loadImage("hayamasq.jpg");
		isPictureDemo=true; isMapDemo=false;
	}

	private void loadImage(String imagefile){
		Toolkit toolkit =  Toolkit.getDefaultToolkit();
		pictureImage = toolkit.getImage(imagefile);
		toolkit.prepareImage(pictureImage,-1,-1,this);
	}
	
	public boolean imageUpdate(Image img , int flg , int x , int y , int w , int h) {
//		when image is loaded, copy it to buffer graphics/image
		if ((flg & ALLBITS) == 0) return true;
		else  {
			//image is ready, then get size and display it
			//pictureSizeX=1280; pictureSizeY=1280; 
			pictureSizeX=pictureImage.getWidth(this);
			pictureSizeY=pictureImage.getHeight(this);
			//prepare buffer image and graphics	
			bufferImage = createImage(pictureSizeX ,pictureSizeY ); //scroll buffer image	
			bufferGraphics=bufferImage.getGraphics();
			bufferGraphics.drawImage
			(pictureImage,0,0,pictureSizeX,pictureSizeY,null);
			
			repaint();
			return false;
		}
	}

	private void rotatePicture(boolean isClockWise) {
		//rotate image CC or CCW
		int[] pix;
		int[] newPix = new int[pictureSizeX * pictureSizeY];
		if(pictureSizeX != pictureSizeY) return; //only for square iamge
		//get pixel matrix to rotate
		PixelGrabber pixcelGrabber = new PixelGrabber(pictureImage,0,0,pictureSizeX,pictureSizeY,true);
		try{ pixcelGrabber.grabPixels(); }catch(InterruptedException e){System.out.println("pixel error");}
		pix = (int[])pixcelGrabber.getPixels();
		//rotate +90 or -90 degree
		if(isClockWise) {
			for(int line=0; line<pictureSizeY; line++)
			for(int column=0; column<pictureSizeX; column++) {
				newPix[column * pictureSizeY  + pictureSizeY - line -1] = pix[line*pictureSizeX + column];
			}
//			adjust scroll posistion to display same place
			//location of picture shown in the center of window is
			int width=this.getWidth(); int height=this.getHeight();
			int centerX = (int)((width/2-offsetX)/this.zoomRatio);
			int centerY = (int)((height/2-offsetY)/this.zoomRatio);
			offsetX = -((int)((pictureSizeX - centerY)*this.zoomRatio) - width/2 );
			offsetY = -((int)(centerX*this.zoomRatio) - height/2);
		}
		else {
			for(int line=0; line<pictureSizeY; line++)
			for(int column=0; column<pictureSizeX; column++) {
				newPix[(pictureSizeX - column -1) * pictureSizeY + line] = pix[line*pictureSizeX + column];
			}
//			adjust scroll posistion to display same place
			//location of picture shown in the center of window is
			int width=this.getWidth(); int height=this.getHeight();
			int centerX = (int)((width/2-offsetX)/this.zoomRatio);
			int centerY = (int)((height/2-offsetY)/this.zoomRatio);
			offsetX = -((int)(centerY*this.zoomRatio) - width/2);
			offsetY = -((int)((pictureSizeY - centerX)*this.zoomRatio) - height/2 );
		}
		//make new image instance from the rotated pixel matrix
		MemoryImageSource ms = new MemoryImageSource(pictureSizeX,pictureSizeY,newPix,0,pictureSizeX);
		pictureImage = createImage(ms);
		bufferGraphics.drawImage(pictureImage,0,0,pictureSizeX,pictureSizeY,null);
		this.repaint();
	}

	private void zoomTo(double newRatio, int centerX, int centerY) { 
	//zoom in/out to the new ratio, by keeping centerXY in window coordinate
		double oldRatio=zoomRatio;
		if(oldRatio == newRatio) return; //no change, do nothing.
		zoomRatio = newRatio;
		offsetX = -(int)((-offsetX + centerX )*newRatio/oldRatio - centerX);
		offsetY = -(int)((-offsetY + centerY )*newRatio/oldRatio - centerY);
		this.repaint();
	}
	
	private void zoomTo(double newRatio) { 
	//zoom in/out to the new ratio, by keeping center in the window
		this.zoomTo
		(newRatio, this.getSize().height / 2, this.getSize().width / 2);
	}
	
//listener methods
	public void actionPerformed(ActionEvent e) {
		if(e.getActionCommand().equals("Black")) currentColor=Color.black;
		else if(e.getActionCommand().equals("Red")) currentColor=Color.red;
		else if(e.getActionCommand().equals("Green")) currentColor=Color.green;
		else if(e.getActionCommand().equals("Blue")) currentColor=Color.blue;
		else if(e.getActionCommand().equals("Yellow")) currentColor=Color.yellow;
		else if(e.getActionCommand().equals("White")) currentColor=Color.white;
		else if(e.getActionCommand().equals("Rotate CW")){this.rotatePicture(true);}
		else if(e.getActionCommand().equals("Rotate CCW")){this.rotatePicture(false);}
		else if(e.getActionCommand().equals("Zoom x0.1"))this.zoomTo(0.1);
		else if(e.getActionCommand().equals("Zoom x0.5"))this.zoomTo(0.5);
		else if(e.getActionCommand().equals("Zoom x1"))this.zoomTo(1.0);
		else if(e.getActionCommand().equals("Zoom x2"))this.zoomTo(2.0);
		else if(e.getActionCommand().equals("Zoom x4"))this.zoomTo(4.0);
		else if(e.getActionCommand().equals("MapDemo")){
			if(this.isPictureDemo) {
				this.loadImage("ochamap.jpg");
				this.isMapDemo=true;this.isPictureDemo=false;
			}
		}
		else if(e.getActionCommand().equals("PictureDemo")){
			if(this.isMapDemo){
				this.loadImage("hayamasq.jpg");
				this.isMapDemo=false;this.isPictureDemo=true;
			}
		}
		else if(e.getActionCommand().equals("Doodle"))this.isDoodleDemo=true;
		else if(e.getActionCommand().equals("ZoomRotate"))this.isDoodleDemo=false;		
	}
	public void mouseClicked(MouseEvent e) {}
	public void mousePressed(MouseEvent e) {
		pressedX=e.getX(); pressedY=e.getY(); 
		lastX=pressedX; lastY=pressedY;
		isPenDown=true;
		if(this.isPaperweightOn) this.wasPaperweightOnWhenPressed=true;
		else this.wasPaperweightOnWhenPressed=false;
	}
	public void mouseReleased(MouseEvent e) {
		isPenDown=false;
		if(this.isDoodleDemo) return; //do nothing anymore
		if(false==this.isPaperweightOn) return; //do nothing anymore
		if(false==this.wasPaperweightOnWhenPressed) return; //do nothing
		int releasedX = e.getX(); 
		int releasedY = e.getY();
		int dx = releasedX - this.pressedX; 
		int dy = releasedY - this.pressedY;
		int threshold = (this.SIZEX + this.SIZEY)/2/10; //10% of screensize
		if((dx*dx+dy*dy) < threshold * threshold) return;
		if(dx==0) dx=1; if(dy==0) dy=1; //just to avoid division by zero
		float tangent=(float)dy / (float)dx;
		/*
		if((tangent*tangent)>1.0) { //it moved up or down
			if(dy>0) { //it moved down
				double newRatio=this.zoomRatio*
				(this.SIZEY-releasedY)/(this.SIZEY-this.pressedY);
				if(newRatio<0.1) return; //too small, do nothing
				this.zoomTo(newRatio,this.pressedX, 
						(this.pressedY+this.getSize().height)/2);
			}
			else { //it moved up
				double newRatio=this.zoomRatio*
				(this.SIZEY-releasedY)/(this.SIZEY-this.pressedY);
				if(newRatio>10) return; //too big, do nothing
				this.zoomTo(newRatio,this.pressedX, 
						(this.pressedY+this.getSize().height)/2);
			}
		}//end of if tangent^2 > 1.0
		*/
		if((tangent*tangent)<1.0)  { //it moved left or right
			if(dx>0) { //it moved right
				this.rotatePicture(true);
			}
			else { //it moved left
				this.rotatePicture(false);
			}
		}//end of else
		this.repaint();
	}
	
	private void stretchTo(int x, int y){
		if(this.isDoodleDemo) return; //do nothing anymore
		if(false==this.isPaperweightOn) return; //do nothing anymore
		if(false==this.wasPaperweightOnWhenPressed) return; //do nothing
		int dx = x - this.pressedX; 
		int dy = y - this.pressedY;
		//int threshold = (this.SIZEX + this.SIZEY)/2/10; //10% of screensize
		//if((dx*dx+dy*dy) < threshold * threshold) return;//too small move
		if(dx==0) dx=1; if(dy==0) dy=1; //just to avoid division by zero
		float tangent=(float)dy / (float)dx;
		if((tangent*tangent)<1.0) return; //movement was not up/down 		
		//it moved up or down
		if(dy>0) { //it moved down
			double newRatio=this.zoomRatio*
			(this.SIZEY-y)/(this.SIZEY-this.lastY);
			if(newRatio<0.1) return; //too small, do nothing
			this.zoomTo(newRatio,this.pressedX, 
					(this.pressedY+this.getSize().height)/2);
		}
		else { //it moved up
			double newRatio=this.zoomRatio*
			(this.SIZEY-y)/(this.SIZEY-this.lastY);
			if(newRatio>10) return; //too big, do nothing
			this.zoomTo(newRatio,this.pressedX, 
					(this.pressedY+this.getSize().height)/2);
		}
	}
	
	private void scrollTo(int x, int y){
		if(this.isPaperweightOn)return;
		offsetX += x - lastX;
		offsetY += y - lastY;
	}
	
	private void doodleTo(int x, int y){
		if(false==this.isPaperweightOn) return;
		if(false==this.isPenDown) return;
		if(false==this.isDoodleDemo) return;
		bufferGraphics.setColor(currentColor);
		bufferGraphics.drawLine
		((int)((lastX-offsetX) / zoomRatio), (int)((lastY-offsetY) / zoomRatio),	
		(int)((x-offsetX) / zoomRatio), (int)((y-offsetY) / zoomRatio) );
	}
	
	public void mouseEntered(MouseEvent e) {}
	public void mouseExited(MouseEvent e) {}
	public void mouseDragged(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();
		this.scrollTo(x,y);//scroll if needed
		this.doodleTo(x,y);//doodle if needed
		this.stretchTo(x,y);//zoom in/out if needed		
		lastX=x; lastY=y;
		repaint();
	}
	public void mouseMoved(MouseEvent e) {}
	public void mouseWheelMoved(MouseWheelEvent e) {
		offsetY += e.getWheelRotation()*5;
		repaint();
	}
	public void keyTyped(KeyEvent e) {}
	public void keyPressed(KeyEvent e) {
		if(KeyEvent.VK_SHIFT == e.getKeyCode()) isPaperweightOn=true;
	}
	public void keyReleased(KeyEvent e) {
		if(KeyEvent.VK_SHIFT == e.getKeyCode()) isPaperweightOn=false;
	}
	public void windowOpened(WindowEvent e) {}
	public void windowClosing(WindowEvent e) {System.exit(0);}
	public void windowClosed(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowActivated(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}

	public void paint(Graphics g) {
		int height=getSize().height;
		int width=getSize().width;
		//check scroll boundary
		if(offsetX < -(pictureSizeX * zoomRatio - width)) offsetX=-(int)(pictureSizeX * zoomRatio - width);
		if(offsetY < -(pictureSizeY * zoomRatio - height)) offsetY=-(int)(pictureSizeY * zoomRatio - height);
		if(offsetX > 0) offsetX=0;
		if(offsetY > 0) offsetY=0;
		/*
		if(offsetX < -(pictureSizeX * zoomRatio )) offsetX=-(int)(pictureSizeX * zoomRatio );
		if(offsetY < -(pictureSizeY * zoomRatio )) offsetY=-(int)(pictureSizeY * zoomRatio );
		if(offsetX > width) offsetX=width;
		if(offsetY > height) offsetY=height;
		if(bufferImage!=null)
		*/
			g.drawImage
			(bufferImage,offsetX,offsetY,
			(int)(pictureSizeX * zoomRatio),(int)(pictureSizeY * zoomRatio),null);
		//System.out.println(offsetX  + ", " + offsetY);
		super.paint(g);
	}

	public static void main(String[] args) {
		new PictViewer();
	}
}
